function fmtNum(num, ccy, decPlaces=2, suffix='', prefix='', zeroIfNull=false) {
    let params = {'data': {'ccy': ccy}, 'value': num};
    let colFmt = {
        'valueType': 'number',
        'decPlaces': decPlaces,
        'ccyColumn': 'ccy',
        'suffix': suffix,
        'prefix': prefix,
        'zeroIfNull': zeroIfNull
    };
    return agGridNumberValueFormatter(params, colFmt);
}

function fmtNumAccounting(num, ccy, decPlaces=0, suffix='', prefix='', zeroIfNull=false) {
    return `${num < 0 ? '(' : ''}${fmtNum(num, ccy, decPlaces, suffix, prefix, zeroIfNull).slice(num < 0 ? 1 : 0)}${num < 0 ? ')' : ''}`;
}

function UIUtils_rgb2hex(r, g, b) {
    let r_ = r.toString(16).length === 1 ? "0" + r.toString(16) : r.toString(16);
    let g_ = g.toString(16).length === 1 ? "0" + g.toString(16) : g.toString(16);
    let b_ = b.toString(16).length === 1 ? "0" + b.toString(16) : b.toString(16);
    return `#${r_}${g_}${b_}`
}

function UIUtils_appendHtml(obj, body) {
    obj.append(body);
    UIUtils_dynamicRendering_onHtmlRendered(obj);
}

function UIUtils_setHtml(obj, body) {
    obj.html(body);
    UIUtils_dynamicRendering_onHtmlRendered(obj);
}

// region dynamic HTML rendering
function UIUtils_dynamicRendering_onHtmlRendered(obj) {
    let newObjs = obj.find('*');

    // disable write-access objects for non-admins
    UIUtils_dynamicRendering_verifyWriteAccess(newObjs);

    // init tabs (if exists)
    UIUtils_dynamicRendering_initTabs(newObjs);

    // init enum select objs
    UIUtils_dynamicRendering_initEnumSelectObjs(newObjs);

    // init select objs with meta modifying callbacks (if exists)
    UIUtils_dynamicRendering_initSelectObjsWithDealMetaModifyingCallbacks(newObjs);

    // init editable data items (if exists)
    UIUtils_dynamicRendering_initEditableDataItems(newObjs);

    // re-listen to escape pressed events
    UIUtils_dynamicRendering_listenToKeyPressedEvent();
}

function UIUtils_dynamicRendering_listenToKeyPressedEvent() {
    $('*')
        .off('keyup')
        .keyup(function (event) {
            if (event.key === 'Escape') {
                console.log('escape pressed');
                genericModal_onEscapePressed(event);
                dynamicModal_onEscapePressed(event);
                editableDataItemModal_onEscapePressed(event);
            }
        });
}

function UIUtils_dynamicRendering_verifyWriteAccess(elems) {
    elems
        .find('.write-access')
        .prop('disabled', $viewMode !== 'admin');
}

function UIUtils_dynamicRendering_initGridColumnCheckboxes(elems) {
    elems.filter('div.grid-column-checkbox').each(function() {

    });
}

function UIUtils_dynamicRendering_initTabs(elems) {
    elems.filter('div.my-nav-tab').each(function () {
        let tabDiv = $(this);
        let tabName = tabDiv.text();
        let defaultTabName = tabDiv.parent().attr('data-default-tab');
        if (tabName === defaultTabName) { tabDiv.removeClass('my-nav-tab').addClass('my-nav-tab-active'); }
        // register click event
        tabDiv.off('click').click(function (e) {
            let clickedTab = $(this);
            let clickedTabName = clickedTab.text();
            let tabOnClickFuncName = clickedTab.attr('data-tab-on-click-func');
            let isClickedTabActive = clickedTab.hasClass('my-nav-tab-active');
            // invoke tab on click
            if (!isClickedTabActive && tabOnClickFuncName) {
                funcUtils_executeByName(tabOnClickFuncName, [clickedTabName]);
            }
            // remove previous active
            clickedTab.parent().find('div.my-nav-tab-active').each(function () {
                $(this).removeClass('my-nav-tab-active').addClass('my-nav-tab');
            });
            // add active to clicked tab
            clickedTab.removeClass('my-nav-tab').addClass('my-nav-tab-active');
        });
    });
}

function UIUtils_dynamicRendering_initEnumSelectObjs(elems) {
    elems.filter('select[data-enum-key]').each(function () {
        selectObj_populateFromEnums($(this));
    });
}

function UIUtils_dynamicRendering_initSelectObjsWithDealMetaModifyingCallbacks(elems) {
    elems
        .filter('select[data-column-key]')
        .each(function () {
            let colKey = $(this).attr('data-column-key');
            let isModifyingObj = $(this).attr('data-is-meta-modifier') === 'true' //todo: delete data-is-meta-modifier
            let modifierCallbackFuncName = $(this).attr('data-meta-modifying-callback');
            if (colKey && isModifyingObj && modifierCallbackFuncName) {
                // console.log(`registering select obj of ${colKey} with modifying callback func ${modifierCallbackFuncName}`);
                $(this)
                    .off('change')
                    .change(function () {
                        funcUtils_executeByName(modifierCallbackFuncName, [colKey, $(this).val()]);
                    });
            }
        });
}

function UIUtils_dynamicRendering_initEditableDataItems(elems) {
    elems
        .filter('span.editable-data-item')
        .off('click')
        .click(function (e) {
            let spanObj = $(this);
            let colKey = spanObj.attr('data-column-key');
            let onEditCallback = spanObj.attr('data-editable-data-item-callback');

            // region checks
            if (!colKey) {
                console.log(`editable data item has no data-column-key defined`);
                return;
            }
            if (colKey in $enum) {
                console.log(`editable data item modal not supported for enum cols (${colKey})`);
                return;
            }
            if (!onEditCallback) {
                console.log(`editable data item modal has no data-editable-data-item-callback defined (${colKey})`);
                return;
            }
            // endregion

            genericModal_hideModal(); // hidden, but still open (will still tick)
            editableDataItemModal(spanObj);
        });
}
// endregion

function setValToObj(obj, val) {
    if (obj.is('span')) { obj.text(val); }
    else if (obj.is('input') && obj.attr('type') === 'date') { obj.val(fmtDateForDateInput(val)); }
    else { obj.val(val); }
    return obj;
}

function getObjVal(obj) {
    if (obj.is('span')) { return obj.text(); }
    if (obj.attr('type') === 'checkbox') { return  obj.is(":checked"); }
    return obj.val();
}

function scrollToBottom(divId) {
    let objDiv = document.getElementById(divId);
	objDiv.scrollTop = objDiv.scrollHeight;
}

function scrollToTop(divId) {
    let objDiv = document.getElementById(divId);
	objDiv.scrollTop = 0;
}

function getDateValueFromInput(obj) {
    let dateStr = obj.val();
    if (!isValidDateStr(dateStr)) {
        obj.css('border', 'solid 2px red');
        return [false, null];
    }
    obj.css('border', '');
    return [true, dateStr];
}

function getFloatValueFromInput(obj) {
    let value = obj.val() * 1;
    if (isNaN(value)) {
        obj.css('border', 'solid 2px red');
        return [false, null];
    }
    obj.css('border', '');
    return [true, value];
}

function blinkBorder(obj, delay=1000) {
    obj.css('border', 'solid 2px red');
    setTimeout(() => obj.css('border', ''), delay);
}

function UIUtils_statusSpan_flash(obj, txt, delay=1000) {
    obj.text(txt);
    setTimeout(() => obj.text(''), delay);
}

function blinkObj(obj, delay=1000) {
    obj.css('transition', 'background-color 0.8s linear');
    let originalColor = obj.css('background-color') || 'transparent';
    obj.css('background-color', '#202020');
    setTimeout(() => obj.css('background-color', originalColor), delay);
}

function ccy2symbol(ccy) {
	return {
		'USD': '$',
		'EUR': '€',
		'GBP': '£',
		'HKD': 'HK$',
		'CHF': 'CHf',
		'CAD': 'C$',
		'AUD': 'A$',
        'BTC': '₿',
        'ETH': 'Ξ',
	}[ccy] || '';
}

function num2bgColor(num) {
    if (isNaN(num)) { return 'red'; }
    if (Math.abs(num) < 1e-3) { return 'gray'; }
    if (num >= 0) { return 'green'; }
    return 'red';
}

function num2color(num) {
    if (isNaN(num)) { return 'red'; }
    if (Math.abs(num) < 1e-2) { return 'gray'; }
    if (num >= 0) { return 'lawngreen'; }
    return 'red';
}

function iso31662ccySymbol(iso3166) {
    switch (iso3166) {
        case $enum.ISO3166Enum.US: return '$';
        case $enum.ISO3166Enum.CA: return 'C$';
        case $enum.ISO3166Enum.AU: return 'A$';
        case $enum.ISO3166Enum.IT: return '€';
        case $enum.ISO3166Enum.DE: return '€';
        default: return '';
    }
}

// region progress span
let $progressSpansAtWork = new Set();

function UIUtils_progressSpan_attachTo(obj, txt='') {
    let spanId = strUtils_randomStr(16);
    // return obj.replaceWith(`<span id="${spanId}">${txt}</span>`);
    obj.html(`<span id="${spanId}">${txt}</span>`);
    return $('#'+spanId);
}

function UIUtils_progressSpan_isWorking(obj) {
    return $progressSpansAtWork.has(obj.attr('id'));
}

function UIUtils_progressSpan_stop(obj, label=null, color=null) {
    $progressSpansAtWork.delete(obj.attr('id'));
    if (label !== null) { obj.text(label); }
    if (color) { obj.css('color', color); }
}

function UIUtils_progressSpan_continue(obj, label, color=null, numDots=3, freqMs=600) {
    if (!UIUtils_progressSpan_isWorking(obj)) {
        UIUtils_progressSpan_start(obj, label, color, numDots, freqMs);
    }
}

function UIUtils_progressSpan_start(obj, label, color=null, numDots=3, freqMs=600) {
    $progressSpansAtWork.add(obj.attr('id'));
    if (color) { obj.css('color', color); }
    function _loop(txt) {
        // halting conditions
        if (!UIUtils_progressSpan_isWorking(obj)) { return; }

        obj.text(txt);
        let currTailSize = strUtils_trailingSize(txt, '.'); // 0 <= currTailSize <= numDots
        let newTailSize = Math.max(1, (currTailSize + 1) %  (numDots + 1))
        let newTail = strUtils_char2str('.', newTailSize);
        setTimeout(() => _loop(label + newTail), freqMs);
    }
    _loop(label);
}

function UIUtils_progressSpan_startMany(objs, label, color=null, numDots=3, freqMs=600) {
    objs.each(function () { UIUtils_progressSpan_start($(this), label, color, numDots, freqMs); });
}

function UIUtils_progressSpan_stopMany(objs, label=null, color=null) {
    objs.each(function () { UIUtils_progressSpan_stop($(this), label, color); });
}

function UIUtils_progressSpan_attachToClasses(objs, classNameFilterFunc=null) {
    let class2objs = UIUtils_class2objs(objs, classNameFilterFunc);
    let class2progressSpanObj = dictUtils_create(Object.keys(class2objs), []);
    Object.keys(class2objs).forEach(className => {
        class2objs[className].forEach(obj => {
            let progressSpanObj = UIUtils_progressSpan_attachTo(obj);
            class2progressSpanObj[className].push(progressSpanObj);
        });
    })
    return class2progressSpanObj;
}
// endregion

function UIUtils_class2objs(objs, classNameFilterFunc=null) {
    classNameFilterFunc = classNameFilterFunc || funcUtils_truthFunction;

    let class2objs = {};
    objs.find('*').each(function () {
        let classList = ($(this).attr('class') || '').split(' ');
        classList.forEach(className => {
            if (classNameFilterFunc(className)) {
                if (!(className in class2objs)) { class2objs[className] = [];}
                class2objs[className].push($(this));
            }
        });
    });
    return class2objs
}

function UIUtils_getAllClassNames(objs, classNameFilterFunc=null) {
    return Object.keys(UIUtils_class2objs(objs, classNameFilterFunc));
}

function UIUtils_filterByClasses(objs, classNames) {
    let classNamesWithDots = arrayUtils_apply(classNames, null, null, x => `.${x}`);
    return objs.find(classNamesWithDots.join(', '));
}